from datetime import datetime
from typing import Optional

from fastapi import APIRouter, Body, HTTPException, Request, Query
from sqlalchemy import case
from sqlalchemy.orm import load_only, selectinload
from sqlmodel import Session, select

from pixlvault.db_models import Picture, PictureStack
from pixlvault.picture_scoring import (
    fetch_smart_score_data,
    get_smart_score_penalised_tags_from_request,
    prepare_smart_score_inputs,
)
from pixlvault.utils.quality.smart_score_utils import SmartScoreUtils
from pixlvault.pixl_logging import get_logger
from pixlvault.utils import safe_model_dict

logger = get_logger(__name__)


def create_router(server) -> APIRouter:
    router = APIRouter()

    def _ensure_secure_when_required(request: Request):
        server.auth.ensure_secure_when_required(request)

    def _normalize_picture_ids(raw_ids) -> list[int]:
        if not isinstance(raw_ids, list):
            raise HTTPException(status_code=400, detail="picture_ids must be a list")
        ids = []
        for raw_id in raw_ids:
            try:
                ids.append(int(raw_id))
            except (TypeError, ValueError):
                raise HTTPException(
                    status_code=400,
                    detail="picture_ids must be integers",
                )
        if not ids:
            raise HTTPException(status_code=400, detail="picture_ids must not be empty")
        return ids

    def _fetch_stack_pictures(session: Session, stack_id: int):
        stack_position_order = case(
            (Picture.stack_position.is_(None), 1),
            else_=0,
        )
        return session.exec(
            select(Picture)
            .where(Picture.stack_id == stack_id)
            .order_by(stack_position_order, Picture.stack_position, Picture.id)
        ).all()

    def _stack_order_key(pic, smart_score_by_id: dict[int, float]):
        score = pic.score or 0
        smart_score = smart_score_by_id.get(pic.id, 0.0)
        created_at = pic.created_at or datetime.min
        created_ts = created_at.timestamp() if isinstance(created_at, datetime) else 0.0
        return (-score, -smart_score, -created_ts, int(pic.id or 0))

    def _compute_smart_score_map(
        request: Request,
        picture_ids: list[int],
    ) -> dict[int, float]:
        if not picture_ids:
            return {}
        try:
            penalised_tags = get_smart_score_penalised_tags_from_request(
                server, request
            )
            good_anchors, bad_anchors, candidates = fetch_smart_score_data(
                server,
                None,
                candidate_ids=picture_ids,
                penalised_tags=penalised_tags,
            )
            if candidates:
                good_list, bad_list, cand_list, cand_ids = prepare_smart_score_inputs(
                    good_anchors,
                    bad_anchors,
                    candidates,
                )
                if cand_list:
                    scores = SmartScoreUtils.calculate_smart_score_batch_numpy(
                        cand_list, good_list, bad_list
                    )
                    return {
                        int(pid): float(score)
                        for pid, score in zip(cand_ids, scores)
                        if score is not None
                    }
        except Exception as exc:
            logger.warning("[stacks] Failed to compute smart scores: %s", exc)
        return {}

    def _ensure_stack_positions(
        request: Request,
        stack_id: int,
        pictures: list[Picture],
    ) -> list[Picture]:
        if not pictures:
            return pictures
        if any(pic.stack_position is not None for pic in pictures):
            return sorted(
                pictures,
                key=lambda pic: (
                    pic.stack_position is None,
                    pic.stack_position or 0,
                    int(pic.id or 0),
                ),
            )

        smart_score_by_id = _compute_smart_score_map(
            request,
            [pic.id for pic in pictures if pic.id is not None],
        )
        ordered = sorted(
            pictures,
            key=lambda pic: _stack_order_key(pic, smart_score_by_id),
        )
        ordered_ids = [pic.id for pic in ordered if pic.id is not None]

        def update_positions(
            session: Session, stack_id_value: int, ordered_ids_value: list[int]
        ):
            stack = session.get(PictureStack, stack_id_value)
            if stack is None:
                return
            pics = session.exec(
                select(Picture).where(Picture.stack_id == stack_id_value)
            ).all()
            pic_by_id = {pic.id: pic for pic in pics}
            for idx, pic_id in enumerate(ordered_ids_value):
                pic = pic_by_id.get(pic_id)
                if pic is None:
                    continue
                pic.stack_position = idx
                session.add(pic)
            stack.updated_at = datetime.utcnow()
            session.add(stack)
            session.commit()

        if ordered_ids:
            server.vault.db.run_task(update_positions, stack_id, ordered_ids)

        return ordered

    @router.get(
        "/stacks/{stack_id}",
        summary="Get stack details",
        description="Returns stack metadata and ordered picture ids for a stack.",
    )
    async def get_stack(stack_id: int, request: Request):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        def fetch_stack(session: Session, stack_id: int):
            stack = session.get(PictureStack, stack_id)
            if not stack:
                return None, []
            pictures = _fetch_stack_pictures(session, stack_id)
            return stack, pictures

        stack, pictures = server.vault.db.run_task(fetch_stack, stack_id)
        if not stack:
            raise HTTPException(status_code=404, detail="Stack not found")

        pictures = _ensure_stack_positions(request, stack_id, pictures)

        payload = safe_model_dict(stack)
        payload["picture_ids"] = [pic.id for pic in pictures]
        return payload

    @router.get(
        "/stacks/{stack_id}/pictures",
        summary="List pictures in stack",
        description="Returns ordered picture payloads for a stack using grid or metadata field sets.",
    )
    async def get_stack_pictures(
        stack_id: int,
        request: Request,
        fields: str = Query("grid"),
        include_deleted: bool = Query(False),
    ):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        def fetch_stack_pictures(
            session: Session,
            stack_id_value: int,
            fields_value: str,
            include_deleted_value: bool,
        ):
            stack = session.get(PictureStack, stack_id_value)
            if not stack:
                return None, None

            select_fields = (
                Picture.grid_fields()
                if fields_value == "grid"
                else Picture.metadata_fields()
            )
            query = select(Picture).where(Picture.stack_id == stack_id_value)
            if not include_deleted_value:
                query = query.where(Picture.deleted.is_(False))
            query = query.order_by(Picture.id)

            if select_fields:
                select_fields = list(set(select_fields) | {"id"})
                scalar_attrs = [
                    getattr(Picture, field)
                    for field in Picture.scalar_fields().intersection(select_fields)
                ]
                if scalar_attrs:
                    query = query.options(load_only(*scalar_attrs))
                rel_attrs = [
                    getattr(Picture, field)
                    for field in Picture.relationship_fields().intersection(
                        select_fields
                    )
                ]
                for rel_attr in rel_attrs:
                    query = query.options(selectinload(rel_attr))

            pictures = session.exec(query).all()
            return select_fields, pictures

        select_fields, pictures = server.vault.db.run_task(
            fetch_stack_pictures,
            stack_id,
            fields,
            include_deleted,
        )
        if select_fields is None:
            raise HTTPException(status_code=404, detail="Stack not found")
        pictures = _ensure_stack_positions(request, stack_id, pictures)
        return [
            {field: safe_model_dict(pic).get(field) for field in select_fields}
            for pic in pictures
        ]

    @router.get(
        "/pictures/{picture_id}/stack",
        summary="Get picture's stack",
        description="Returns the stack containing a picture, or null stack information when unstacked.",
    )
    async def get_stack_for_picture(picture_id: int, request: Request):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        def fetch_stack_for_picture(session: Session, picture_id: int):
            pic = session.get(Picture, picture_id)
            if not pic or not pic.stack_id:
                return None, None, []
            stack = session.get(PictureStack, pic.stack_id)
            pictures = _fetch_stack_pictures(session, pic.stack_id)
            return pic.stack_id, stack, pictures

        stack_id, stack, pictures = server.vault.db.run_task(
            fetch_stack_for_picture, picture_id
        )
        if not stack_id or not stack:
            return {"stack_id": None, "picture_ids": []}

        pictures = _ensure_stack_positions(request, stack_id, pictures)

        payload = safe_model_dict(stack)
        payload["picture_ids"] = [pic.id for pic in pictures]
        return payload

    @router.post(
        "/stacks",
        summary="Create stack",
        description="Creates a new stack or reuses an existing compatible one and assigns provided pictures to it.",
    )
    async def create_stack(payload: dict = Body(...), request: Request = None):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        picture_ids = _normalize_picture_ids(payload.get("picture_ids") or [])
        name = payload.get("name")
        if name is not None and not isinstance(name, str):
            name = str(name)

        def create_or_assign_stack(
            session: Session,
            picture_ids: list[int],
            name: Optional[str],
        ) -> int:
            pictures = session.exec(
                select(Picture).where(Picture.id.in_(picture_ids))
            ).all()
            if len(pictures) != len(picture_ids):
                missing = sorted(set(picture_ids) - {pic.id for pic in pictures})
                raise HTTPException(
                    status_code=404,
                    detail=f"Pictures not found: {missing}",
                )

            existing_stack_ids = {pic.stack_id for pic in pictures if pic.stack_id}
            if len(existing_stack_ids) > 1:
                raise HTTPException(
                    status_code=409,
                    detail="Pictures already belong to multiple stacks",
                )

            if existing_stack_ids:
                stack_id = existing_stack_ids.pop()
                stack = session.get(PictureStack, stack_id)
                if stack is None:
                    raise HTTPException(status_code=404, detail="Stack not found")
            else:
                stack = PictureStack(name=name)
                session.add(stack)
                session.commit()
                session.refresh(stack)

            existing_positions = []
            if stack.id is not None:
                rows = session.exec(
                    select(Picture.stack_position).where(
                        Picture.stack_id == stack.id,
                        Picture.stack_position.is_not(None),
                    )
                ).all()
                existing_positions = [row for row in rows if row is not None]
            next_position = max(existing_positions) + 1 if existing_positions else None

            for pic in pictures:
                pic.stack_id = stack.id
                if next_position is not None and pic.stack_position is None:
                    pic.stack_position = next_position
                    next_position += 1
                session.add(pic)

            stack.updated_at = datetime.utcnow()
            session.add(stack)
            session.commit()
            if stack.id is None:
                raise HTTPException(status_code=500, detail="Failed to create stack")
            return stack.id

        def fetch_stack_payload(session: Session, stack_id_value: int) -> dict:
            stack = session.get(PictureStack, stack_id_value)
            if stack is None:
                raise HTTPException(status_code=404, detail="Stack not found")
            return safe_model_dict(stack)

        stack_id = server.vault.db.run_task(create_or_assign_stack, picture_ids, name)
        pictures = server.vault.db.run_task(_fetch_stack_pictures, stack_id)
        pictures = _ensure_stack_positions(request, stack_id, pictures)
        payload = server.vault.db.run_task(fetch_stack_payload, stack_id)
        payload["picture_ids"] = [pic.id for pic in pictures]
        return payload

    @router.patch(
        "/stacks/{stack_id}/order",
        summary="Reorder stack",
        description="Sets explicit order for all members in a stack using a complete ordered id list.",
    )
    async def reorder_stack(
        stack_id: int, payload: dict = Body(...), request: Request = None
    ):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        picture_ids = _normalize_picture_ids(payload.get("picture_ids") or [])
        unique_ids = list(dict.fromkeys(picture_ids))
        if len(unique_ids) != len(picture_ids):
            raise HTTPException(status_code=400, detail="picture_ids must be unique")

        def update_stack_order(
            session: Session, stack_id_value: int, ordered_ids: list[int]
        ):
            stack = session.get(PictureStack, stack_id_value)
            if stack is None:
                return None

            pics = session.exec(
                select(Picture).where(Picture.stack_id == stack_id_value)
            ).all()
            pic_by_id = {pic.id: pic for pic in pics}
            stack_ids = set(pic_by_id.keys())
            if stack_ids != set(ordered_ids):
                raise HTTPException(
                    status_code=400,
                    detail="picture_ids must include every picture in the stack",
                )

            for idx, pic_id in enumerate(ordered_ids):
                pic = pic_by_id.get(pic_id)
                if pic is None:
                    continue
                pic.stack_position = idx
                session.add(pic)

            stack.updated_at = datetime.utcnow()
            session.add(stack)
            session.commit()
            return ordered_ids

        result = server.vault.db.run_task(update_stack_order, stack_id, unique_ids)
        if result is None:
            raise HTTPException(status_code=404, detail="Stack not found")
        return {"stack_id": stack_id, "picture_ids": result}

    @router.post(
        "/stacks/{stack_id}/members",
        summary="Add stack members",
        description="Adds pictures to an existing stack while preventing cross-stack membership conflicts.",
    )
    async def add_stack_members(
        stack_id: int, payload: dict = Body(...), request: Request = None
    ):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        picture_ids = _normalize_picture_ids(payload.get("picture_ids") or [])

        def add_members(session: Session, stack_id: int, picture_ids: list[int]):
            stack = session.get(PictureStack, stack_id)
            if stack is None:
                raise HTTPException(status_code=404, detail="Stack not found")

            pictures = session.exec(
                select(Picture).where(Picture.id.in_(picture_ids))
            ).all()
            if len(pictures) != len(picture_ids):
                missing = sorted(set(picture_ids) - {pic.id for pic in pictures})
                raise HTTPException(
                    status_code=404,
                    detail=f"Pictures not found: {missing}",
                )

            conflicts = [
                pic.id for pic in pictures if pic.stack_id not in (None, stack_id)
            ]
            if conflicts:
                raise HTTPException(
                    status_code=409,
                    detail=f"Pictures already in another stack: {sorted(conflicts)}",
                )

            existing_positions = []
            rows = session.exec(
                select(Picture.stack_position).where(
                    Picture.stack_id == stack_id,
                    Picture.stack_position.is_not(None),
                )
            ).all()
            existing_positions = [row for row in rows if row is not None]
            next_position = max(existing_positions) + 1 if existing_positions else None

            for pic in pictures:
                pic.stack_id = stack_id
                if next_position is not None and pic.stack_position is None:
                    pic.stack_position = next_position
                    next_position += 1
                session.add(pic)

            stack.updated_at = datetime.utcnow()
            session.add(stack)
            session.commit()
            return stack

        stack = server.vault.db.run_task(add_members, stack_id, picture_ids)
        pictures = server.vault.db.run_task(_fetch_stack_pictures, stack_id)
        pictures = _ensure_stack_positions(request, stack_id, pictures)
        payload = safe_model_dict(stack)
        payload["picture_ids"] = [pic.id for pic in pictures]
        return payload

    @router.delete(
        "/stacks/{stack_id}/members",
        summary="Remove stack members",
        description="Removes pictures from a stack and deletes the stack when one or fewer members remain.",
    )
    async def remove_stack_members(
        stack_id: int, payload: dict = Body(...), request: Request = None
    ):
        _ensure_secure_when_required(request)
        server.auth.require_user_id(request)

        picture_ids = _normalize_picture_ids(payload.get("picture_ids") or [])

        def remove_members(session: Session, stack_id: int, picture_ids: list[int]):
            stack = session.get(PictureStack, stack_id)
            if stack is None:
                raise HTTPException(status_code=404, detail="Stack not found")

            pictures = session.exec(
                select(Picture).where(Picture.id.in_(picture_ids))
            ).all()
            for pic in pictures:
                if pic.stack_id == stack_id:
                    pic.stack_id = None
                    session.add(pic)

            remaining = session.exec(
                select(Picture).where(Picture.stack_id == stack_id)
            ).all()

            if len(remaining) <= 1:
                for pic in remaining:
                    pic.stack_id = None
                    pic.stack_position = None
                    session.add(pic)
                session.delete(stack)
                session.commit()
                return None

            stack.updated_at = datetime.utcnow()
            session.add(stack)
            session.commit()
            return stack

        stack = server.vault.db.run_task(remove_members, stack_id, picture_ids)
        if stack is None:
            return {"status": "success", "stack_id": None, "picture_ids": picture_ids}

        payload = safe_model_dict(stack)
        payload["picture_ids"] = picture_ids
        return payload

    return router



"""Built-in blur/sharpen plugin."""

from __future__ import annotations

import os
import tempfile
from typing import Any

import cv2
import numpy as np
from PIL import Image, ImageEnhance, ImageFilter

from pixlvault.image_plugins.base import ImagePlugin


class BlurSharpenPlugin(ImagePlugin):
    name = "blur_sharpen"
    display_name = "Blur / Sharpen"
    description = "Apply blur or sharpen effect to images or videos."
    supports_images = True
    supports_videos = True

    MODES = {"blur", "sharpen"}

    def parameter_schema(self) -> list[dict[str, Any]]:
        return [
            {
                "name": "mode",
                "label": "Mode",
                "type": "string",
                "default": "blur",
                "enum": sorted(self.MODES),
                "description": "Choose whether to blur or sharpen.",
            },
            {
                "name": "strength",
                "label": "Strength",
                "type": "number",
                "default": 1.0,
                "description": "Effect strength (higher means stronger).",
            },
        ]

    def run(
        self,
        images: list[Image.Image],
        parameters: dict[str, Any] | None = None,
        progress_callback=None,
        error_callback=None,
    ) -> list[Image.Image]:
        params = parameters or {}
        mode = str(params.get("mode") or "blur").strip().lower()
        if mode not in self.MODES:
            mode = "blur"
        strength = self._coerce_positive_number(params.get("strength"), 1.0)

        out: list[Image.Image] = []
        total = len(images)
        for idx, image in enumerate(images):
            try:
                filtered = self._apply_mode(image, mode, strength)
                out.append(filtered)
                self.report_progress(
                    progress_callback,
                    current=idx + 1,
                    total=total,
                    message=f"Processed image {idx + 1}/{total}",
                )
            except Exception as exc:
                self.report_error(
                    error_callback,
                    index=idx,
                    message="Failed to apply blur/sharpen",
                    details={"error": str(exc)},
                )
                out.append(image.copy())
        return out

    def run_video(
        self,
        source_path: str,
        parameters: dict[str, Any] | None = None,
        progress_callback=None,
        error_callback=None,
    ) -> tuple[bytes, str]:
        params = parameters or {}
        mode = str(params.get("mode") or "blur").strip().lower()
        if mode not in self.MODES:
            mode = "blur"
        strength = self._coerce_positive_number(params.get("strength"), 1.0)

        cap = cv2.VideoCapture(source_path)
        if not cap.isOpened():
            raise ValueError(f"Failed to open video file: {source_path}")

        frame_total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) or 0
        fps = float(cap.get(cv2.CAP_PROP_FPS) or 0.0)
        if fps <= 0:
            fps = 24.0
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH) or 0)
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT) or 0)
        if width <= 0 or height <= 0:
            cap.release()
            raise ValueError(f"Invalid video dimensions for {source_path}")

        temp_path = ""
        writer = None
        output_ext = ".mp4"
        processed = 0
        try:
            source_ext = os.path.splitext(source_path)[1].lower()
            writer_candidates = [
                (".mp4", "avc1"),
                (".mp4", "H264"),
                (".webm", "VP80"),
                (".webm", "VP90"),
                (".mp4", "mp4v"),
            ]
            if source_ext == ".webm":
                writer_candidates = [
                    (".webm", "VP80"),
                    (".webm", "VP90"),
                    (".mp4", "avc1"),
                    (".mp4", "H264"),
                    (".mp4", "mp4v"),
                ]

            for candidate_ext, codec in writer_candidates:
                with tempfile.NamedTemporaryFile(
                    suffix=candidate_ext,
                    delete=False,
                ) as temp_file:
                    temp_path = temp_file.name
                candidate_writer = cv2.VideoWriter(
                    temp_path,
                    cv2.VideoWriter_fourcc(*codec),
                    fps,
                    (width, height),
                )
                if candidate_writer.isOpened():
                    writer = candidate_writer
                    output_ext = candidate_ext
                    break
                candidate_writer.release()
                try:
                    os.remove(temp_path)
                except OSError:
                    pass
                temp_path = ""

            if writer is None:
                raise ValueError("Failed to open output video writer")

            while True:
                ok, frame = cap.read()
                if not ok or frame is None:
                    break

                rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                filtered = self._apply_mode(Image.fromarray(rgb_frame), mode, strength)
                filtered_bgr = cv2.cvtColor(
                    np.array(filtered.convert("RGB")),
                    cv2.COLOR_RGB2BGR,
                )
                writer.write(filtered_bgr)

                processed += 1
                self.report_progress(
                    progress_callback,
                    current=processed,
                    total=frame_total if frame_total > 0 else processed,
                    message=f"Processed video frame {processed}",
                )

            if processed == 0:
                raise ValueError("No frames processed from video")

            writer.release()
            writer = None
            cap.release()

            with open(temp_path, "rb") as handle:
                return handle.read(), output_ext
        except Exception as exc:
            self.report_error(
                error_callback,
                index=0,
                message="Failed to apply blur/sharpen to video",
                details={"error": str(exc), "source_path": source_path},
            )
            raise
        finally:
            if writer is not None:
                writer.release()
            cap.release()
            if temp_path and os.path.exists(temp_path):
                try:
                    os.remove(temp_path)
                except OSError:
                    pass

    @staticmethod
    def _coerce_positive_number(value: Any, default: float) -> float:
        try:
            parsed = float(value)
        except (TypeError, ValueError):
            return default
        if parsed <= 0:
            return default
        return parsed

    @staticmethod
    def _apply_mode(image: Image.Image, mode: str, strength: float) -> Image.Image:
        rgb = image.convert("RGB")
        if mode == "sharpen":
            factor = 1.0 + (strength * 1.5)
            return ImageEnhance.Sharpness(rgb).enhance(factor)
        radius = max(0.1, strength * 1.2)
        return rgb.filter(ImageFilter.GaussianBlur(radius=radius))



from .task_type import TaskType
from .base_task_finder import BaseTaskFinder, TaskFinderRegistry
from .description_task import DescriptionTask
from .face_extraction_task import FaceExtractionTask
from .face_quality_task import FaceQualityTask
from .quality_task import QualityTask
from .image_embedding_task import ImageEmbeddingTask
from .missing_face_quality_finder import MissingFaceQualityFinder
from .missing_face_extraction_finder import MissingFaceExtractionFinder
from .missing_image_embedding_finder import MissingImageEmbeddingFinder
from .missing_likeness_finder import MissingLikenessFinder
from .missing_likeness_parameters_finder import MissingLikenessParametersFinder
from .missing_description_finder import MissingDescriptionFinder
from .missing_quality_finder import MissingQualityFinder
from .missing_text_embedding_finder import MissingTextEmbeddingFinder
from .missing_tag_finder import MissingTagFinder
from .missing_watch_folder_import_finder import MissingWatchFolderImportFinder
from .tag_task import TagTask
from .text_embedding_task import TextEmbeddingTask
from .likeness_parameters_task import LikenessParametersTask
from .likeness_task import LikenessTask

__all__ = [
    "TaskType",
    "BaseTaskFinder",
    "TaskFinderRegistry",
    "DescriptionTask",
    "FaceQualityTask",
    "FaceExtractionTask",
    "ImageEmbeddingTask",
    "QualityTask",
    "MissingFaceQualityFinder",
    "MissingFaceExtractionFinder",
    "MissingImageEmbeddingFinder",
    "MissingLikenessFinder",
    "MissingLikenessParametersFinder",
    "MissingDescriptionFinder",
    "MissingQualityFinder",
    "MissingTextEmbeddingFinder",
    "MissingTagFinder",
    "MissingWatchFolderImportFinder",
    "TagTask",
    "TextEmbeddingTask",
    "LikenessParametersTask",
    "LikenessTask",
]



import gc
import time
from fastapi.testclient import TestClient

import collections
import tempfile
import os
import json
import psutil
import tracemalloc

from pixlvault.pixl_logging import get_logger
from pixlvault.server import Server
from tests.utils import upload_pictures_and_wait

logger = get_logger(__name__)


def log_resources(label):
    process = psutil.Process()
    rss = process.memory_info().rss / (1024 * 1024)
    logger.info(f"[RESOURCE] {label}: RSS={rss:.2f}MB, Threads={process.num_threads()}")
    logger.info(f"[RESOURCE] {label}: gc objects={len(gc.get_objects())}")
    counter = collections.Counter(type(obj) for obj in gc.get_objects())
    logger.info(f"[RESOURCE] {label}: Top object types: {counter.most_common(5)}")
    if tracemalloc.is_tracing():
        logger.info(
            f"[RESOURCE] {label}: Tracemalloc current={tracemalloc.get_traced_memory()[0] / (1024 * 1024):.2f}MB, peak={tracemalloc.get_traced_memory()[1] / (1024 * 1024):.2f}MB"
        )


def setup_server_with_temp_db():
    log_resources("Before setup_server_with_temp_db")
    temp_dir = tempfile.TemporaryDirectory()
    image_root = os.path.join(temp_dir.name, "images")
    os.makedirs(image_root, exist_ok=True)
    server_config_path = os.path.join(temp_dir.name, "server-config.json")
    with open(server_config_path, "w") as f:
        f.write(json.dumps({"port": 8000}))
    server = Server(server_config_path)
    client = TestClient(server.api)

    resp = client.post(
        "/login", json={"username": "testuser", "password": "testpassword"}
    )
    assert resp.status_code == 200
    return temp_dir, client, server


def test_create_and_list_picture_set():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create a new picture set
        resp = client.post(
            "/picture_sets", json={"name": "TestSet", "description": "A test set"}
        )
        assert resp.status_code == 200
        data = resp.json()
        set_id = data["picture_set"]["id"]
        # List all picture sets
        resp = client.get("/picture_sets")
        assert resp.status_code == 200
        sets = resp.json()
        assert any(s["id"] == set_id for s in sets)
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()


def test_get_picture_set_metadata_and_members():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create a new set
        resp = client.post("/picture_sets", json={"name": "MetaSet"})
        set_id = resp.json()["picture_set"]["id"]
        # Get metadata
        resp = client.get(f"/picture_sets/{set_id}?info=true")
        assert resp.status_code == 200
        meta = resp.json()
        assert meta["id"] == set_id
        # Get members (should be empty)
        resp = client.get(f"/picture_sets/{set_id}/members")
        assert resp.status_code == 200
        assert resp.json()["picture_ids"] == []
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()


def test_add_and_remove_picture_from_set():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create set
        resp = client.post("/picture_sets", json={"name": "AddRemSet"})
        set_id = resp.json()["picture_set"]["id"]
        # Add a real picture from the pictures/ directory
        import glob

        # Find a real PNG in the pictures/ directory
        png_files = glob.glob(
            os.path.join(os.path.dirname(__file__), "..", "pictures", "*.png")
        )
        assert png_files, "No PNG files found in pictures/ directory for test."
        img_path = png_files[0]
        with client.websocket_connect("/ws/updates") as ws:
            import threading
            import queue

            imported = False
            messages = queue.Queue()

            def recv_loop():
                try:
                    while True:
                        messages.put(ws.receive_json())
                except Exception:
                    return

            thread = threading.Thread(target=recv_loop, daemon=True)
            thread.start()

            with open(img_path, "rb") as f:
                files = {"file": (os.path.basename(img_path), f, "image/png")}
                import_status = upload_pictures_and_wait(client, files, timeout_s=120)
            assert import_status["status"] == "completed"
            # Get picture id
            pic_id = import_status["results"][0]["picture_id"]
            # Add to set
            resp = client.post(f"/picture_sets/{set_id}/members/{pic_id}")
            assert resp.status_code == 200
            # PICTURE_IMPORTED was fired during the blocking import call above;
            # the message is already queued.  Reset the deadline so we have a
            # fresh window to drain it.
            deadline = time.time() + 5
            while time.time() < deadline:
                try:
                    payload = messages.get(timeout=0.2)
                except queue.Empty:
                    continue
                if not isinstance(payload, dict):
                    continue
                if payload.get("event") == "PICTURE_IMPORTED" and pic_id in (
                    payload.get("picture_ids") or []
                ):
                    imported = True
                    break
            assert imported, "Timed out waiting for PICTURE_IMPORTED websocket event"
        # Check members
        resp = client.get(f"/picture_sets/{set_id}/members")
        assert pic_id in resp.json()["picture_ids"]
        # Remove from set
        resp = client.delete(f"/picture_sets/{set_id}/members/{pic_id}")
        assert resp.status_code == 200
        resp = client.get(f"/picture_sets/{set_id}/members")
        assert pic_id not in resp.json()["picture_ids"]
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()


def test_update_and_delete_picture_set():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create set
        resp = client.post("/picture_sets", json={"name": "UpdDelSet"})
        set_id = resp.json()["picture_set"]["id"]
        # Update name/description
        resp = client.patch(
            f"/picture_sets/{set_id}", json={"name": "Updated", "description": "Desc"}
        )
        assert resp.status_code == 200
        resp = client.get(f"/picture_sets/{set_id}?info=true")
        meta = resp.json()
        assert meta["name"] == "Updated"
        assert meta["description"] == "Desc"
        # Delete set
        resp = client.delete(f"/picture_sets/{set_id}")
        assert resp.status_code == 200
        resp = client.get(f"/picture_sets/{set_id}?info=true")
        assert resp.status_code == 404
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()


def test_reference_picture_set_created_with_character():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create a character
        char_name = "RefSetChar"
        resp = client.post("/characters", json={"name": char_name})
        assert resp.status_code == 200
        char = resp.json()["character"]
        assert char is not None
        # List all picture sets
        resp = client.get("/picture_sets")
        assert resp.status_code == 200
        sets = resp.json()
        # There should be a reference set with name 'reference_pictures' and description == char_name
        ref_sets = [
            s
            for s in sets
            if s["name"] == "reference_pictures" and s["description"] == char_name
        ]
        assert len(ref_sets) == 1, (
            f"Expected 1 reference set for character, found {len(ref_sets)}"
        )
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()


def test_reference_picture_set_unique_per_character():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create two characters
        resp1 = client.post("/characters", json={"name": "CharA"})
        resp2 = client.post("/characters", json={"name": "CharB"})
        assert resp1.status_code == 200 and resp2.status_code == 200
        # List all picture sets
        resp = client.get("/picture_sets")
        sets = resp.json()
        ref_a = [
            s
            for s in sets
            if s["name"] == "reference_pictures" and s["description"] == "CharA"
        ]
        ref_b = [
            s
            for s in sets
            if s["name"] == "reference_pictures" and s["description"] == "CharB"
        ]
        assert len(ref_a) == 1, "Reference set for CharA missing or duplicated"
        assert len(ref_b) == 1, "Reference set for CharB missing or duplicated"
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()


def test_no_duplicate_reference_picture_sets():
    temp_dir, client, server = setup_server_with_temp_db()
    try:
        # Create a character
        char_name = "NoDupChar"
        resp = client.post("/characters", json={"name": char_name})
        assert resp.status_code == 200
        # List all picture sets
        resp = client.get("/picture_sets")
        sets = resp.json()
        ref_sets = [
            s
            for s in sets
            if s["name"] == "reference_pictures" and s["description"] == char_name
        ]
        assert len(ref_sets) == 1
        # Try to create the same character name again (should create a new character and a new reference set with the same description)
        client.post("/characters", json={"name": char_name})
        # Accept either error or success, and allow multiple reference sets with the same description
        resp = client.get("/picture_sets")
        sets = resp.json()
        ref_sets = [
            s
            for s in sets
            if s["name"] == "reference_pictures" and s["description"] == char_name
        ]
        assert len(ref_sets) >= 1, "No reference picture set found for character name"
    finally:
        server.vault.close()
        temp_dir.cleanup()
        gc.collect()



"""baseline schema

Revision ID: 0001_baseline
Revises:
Create Date: 2026-03-06 00:00:00.000000
"""

from typing import Sequence, Union

from alembic import op
from sqlmodel import SQLModel

# Import all models to register metadata.
import pixlvault.db_models  # noqa: F401


# revision identifiers, used by Alembic.
revision: str = "0001_baseline"  # noqa: F841
down_revision: Union[str, None] = None  # noqa: F841
branch_labels: Union[str, Sequence[str], None] = None  # noqa: F841
depends_on: Union[str, Sequence[str], None] = None  # noqa: F841


def upgrade() -> None:
    SQLModel.metadata.create_all(op.get_bind())


def downgrade() -> None:
    SQLModel.metadata.drop_all(op.get_bind())



from sqlmodel import Session

from .base_task_finder import BaseTaskFinder
from .likeness_task import LikenessTask


class MissingLikenessFinder(BaseTaskFinder):
    """Find pending likeness work and create a LikenessTask."""

    def __init__(self, database):
        super().__init__()
        self._db = database

    def finder_name(self) -> str:
        return "MissingLikenessFinder"

    def find_task(self):
        queue_count, candidate_count, pair_count = self._db.run_immediate_read_task(
            self._likeness_state
        )
        if int(queue_count or 0) > 0:
            return LikenessTask(database=self._db)
        if int(candidate_count or 0) > 0 and int(pair_count or 0) == 0:
            return LikenessTask(database=self._db)
        return None

    @staticmethod
    def _likeness_state(session: Session):
        return (
            LikenessTask.count_queue(session),
            LikenessTask.count_total_candidates(session),
            LikenessTask.count_total_pairs(session),
        )



from pixlvault.work_planner import WorkPlanner


class _FakeTask:
    def __init__(self, task_id: str):
        self.id = task_id


class _OneShotFinder:
    def __init__(self):
        self._returned = False

    def finder_name(self) -> str:
        return "TestFinder"

    def max_inflight_tasks(self) -> int:
        return 1

    def find_task(self):
        if self._returned:
            return None
        self._returned = True
        return _FakeTask("task-1")

    def on_task_complete(self, task, error):
        return None


class _FastCompleteRunner:
    def __init__(self):
        self.on_submit = None

    def submit(self, task):
        if callable(self.on_submit):
            self.on_submit(task)
        return task.id


def test_inflight_decrements_when_task_completes_during_submit():
    runner = _FastCompleteRunner()
    finder = _OneShotFinder()
    planner = WorkPlanner(task_runner=runner, task_finders=[finder])

    runner.on_submit = lambda task: planner.on_task_complete(task, None)

    submitted = planner._run_finders_once()

    assert submitted is True
    assert planner.inflight_count("TestFinder") == 0
    assert planner._finder_by_task_id == {}



from PIL import Image
import io

import gc
import logging
import sys
import time
import tempfile
import os

from pixlvault.db_models import Character, Face, Picture
from sqlmodel import select
from pixlvault.server import Server
from pixlvault.pixl_logging import get_logger
from pixlvault.tasks.task_type import TaskType


logger = get_logger(__name__)

logging.basicConfig(level=logging.INFO, stream=sys.stdout)
logging.info("Debug info")


def wait_for_worker_completion(worker, timeout=60):
    start = time.time()
    while time.time() - start < timeout:
        if not worker.is_alive() or not worker.is_busy():
            return True
        time.sleep(0.5)
    return False


def test_facial_features():
    with tempfile.TemporaryDirectory() as temp_dir:
        image_root = os.path.join(temp_dir, "images")
        os.makedirs(image_root, exist_ok=True)
        server_config_path = os.path.join(temp_dir, "server-config.json")
        with Server(server_config_path) as server:
            server.vault.import_default_data(add_tagger_test_images=True)

            # Check face counts for TaggerTest*.png
            pics = server.vault.db.run_task(lambda session: Picture.find(session))

            futures = []
            for pic in pics:
                logger.info(
                    "Scheduling watch for picture %s with description %s"
                    % (pic.file_path, pic.description)
                )
                futures.append(
                    server.vault.get_worker_future(
                        TaskType.FACE_EXTRACTION, Picture, pic.id, "faces"
                    )
                )

            # Wait for all face detection futures to complete
            results = [future.result(timeout=60) for future in futures]
            assert all(results), "Not all pictures were processed in time"

            # Now run assertions as before
            pics = server.vault.db.run_task(lambda session: Picture.find(session))
            assert len(pics) > 0, "No pictures found in vault"
            for pic in pics:
                if pic.description and pic.description.startswith("TaggerTest"):
                    faces = server.vault.db.run_task(
                        lambda session: Face.find(session, picture_id=pic.id)
                    )
                    # Check face count as before
                    if "Multi" in os.path.basename(pic.description):
                        assert 2 <= len(faces) <= 3, (
                            f"{pic.description} should have 2 or 3 faces, found {len(faces)}"
                        )
                        logger.info(
                            "Picture %s has %d faces as expected"
                            % (pic.description, len(faces))
                        )
                    else:
                        assert len(faces) == 1, (
                            f"{pic.description} should have 1 face, found {len(faces)}"
                        )
                        logger.info(
                            "Picture %s has %d faces as expected"
                            % (pic.description, len(faces))
                        )
                    # New: Check that each face has a non-empty face_bbox
                    for face in faces:
                        assert face.bbox not in (None, "", "null"), (
                            f"Face bbox missing for {pic.description} face_index={face.face_index}"
                        )
                        logger.info(
                            f"{pic.description} face_index={face.face_index} has bbox: {face.bbox}"
                        )
                        assert face.face_index >= 0, (
                            f"Face index should be non-negative for {pic.description} face_index={face.face_index}"
                        )

                        assert face.features is not None, (
                            f"Face features missing for {pic.description} face_index={face.face_index}"
                        )
                        logger.info(
                            f"{pic.description} face_index={face.face_index} has features: {face.features is not None}"
                        )
    gc.collect()


def test_character_thumbnail_endpoint():
    with tempfile.TemporaryDirectory() as temp_dir:
        image_root = os.path.join(temp_dir, "images")
        os.makedirs(image_root, exist_ok=True)
        server_config_path = os.path.join(temp_dir, "server-config.json")
        with Server(server_config_path) as server:
            server.vault.import_default_data(add_tagger_test_images=True)

            # Check face counts for TaggerTest*.png
            pics = server.vault.db.run_task(lambda session: Picture.find(session))

            futures = []
            for pic in pics:
                logger.info(
                    "Scheduling watch for picture %s with description %s"
                    % (pic.file_path, pic.description)
                )
                futures.append(
                    server.vault.get_worker_future(
                        TaskType.FACE_EXTRACTION, Picture, pic.id, "faces"
                    )
                )

            # Wait for all face detection futures to complete
            results = [future.result(timeout=60) for future in futures]
            assert len(results) == len(futures), (
                "Not all pictures were processed in time"
            )

            # Assign the default character to the largest face in each picture
            chars = server.vault.db.run_task(lambda session: Character.find(session))
            char = chars[0]
            pics = server.vault.db.run_task(lambda session: Picture.find(session))
            for pic in pics:
                faces = server.vault.db.run_task(
                    lambda session: Face.find(session, picture_id=pic.id)
                )
                if not faces:
                    continue
                # Find the largest face by area
                largest_face = max(
                    faces, key=lambda f: (f.width or 0) * (f.height or 0)
                )

                def assign_char(session, face_id, char_id):
                    face = session.get(Face, face_id)
                    face.character_id = char_id
                    session.add(face)
                    session.commit()

                server.vault.db.run_task(assign_char, largest_face.id, char.id)

            # Now get the character with faces
            chars = server.vault.db.run_task(
                lambda session: Character.find(session, select_fields=["faces"])
            )
            char = None
            for c in chars:
                if c.faces:
                    char = c
                    break
            assert char is not None, "No character with faces found"

            # Get the thumbnail via the endpoint
            from fastapi.testclient import TestClient

            client = TestClient(server.api)

            resp = client.post(
                "/login", json={"username": "testuser", "password": "testpassword"}
            )
            assert resp.status_code == 200

            response = client.get(f"/characters/{char.id}/thumbnail")
            assert response.status_code == 200, (
                f"Thumbnail endpoint failed: {response.status_code}"
            )
            assert response.headers["content-type"] == "image/png"

            # Load the image from response
            thumb_img = Image.open(io.BytesIO(response.content))
            # Get the best face and crop from the database
            picture_ids = {
                face.picture_id for face in char.faces if face.picture_id is not None
            }

            def fetch_picture_scores(session, ids):
                if not ids:
                    return {}
                rows = session.exec(
                    select(Picture.id, Picture.score).where(Picture.id.in_(ids))
                ).all()
                return {pid: (score or 0) for pid, score in rows}

            score_by_picture_id = server.vault.db.run_task(
                fetch_picture_scores, picture_ids
            )
            best_face = max(
                char.faces,
                key=lambda face: (
                    score_by_picture_id.get(face.picture_id, 0),
                    face.id or 0,
                ),
            )
            # Query the picture for this face (avoid DetachedInstanceError)
            best_pic = server.vault.db.run_task(
                lambda session: session.get(Picture, best_face.picture_id)
            )
            from pixlvault.utils.image_processing.face_utils import FaceUtils

            bbox = best_face.bbox
            logger.info(
                f"Cropping bbox: {bbox} from picture {best_pic.file_path} with description {best_pic.description}"
            )
            crop_img = FaceUtils.crop_face_bbox_exact(
                os.path.join(server.vault.image_root, best_pic.file_path), bbox
            )
            # Save both images for manual inspection
            outdir = os.path.join(
                os.path.dirname(__file__), "..", "tmp", "face_thumbnails"
            )
            os.makedirs(outdir, exist_ok=True)
            thumb_img.save(os.path.join(outdir, f"character_{char.id}_endpoint.png"))
            crop_img.save(os.path.join(outdir, f"character_{char.id}_dbcrop.png"))



from enum import Enum


class TaskType(str, Enum):
    """Identifies each background worker / task-runner lane."""

    FACE_EXTRACTION = "FaceExtractionTask"
    TAGGER = "TagTask"
    QUALITY = "QualityTask"
    FACE_QUALITY = "FaceQualityTask"
    LIKENESS = "LikenessTask"
    LIKENESS_PARAMETERS = "LikenessParametersTask"
    DESCRIPTION = "DescriptionTask"
    TEXT_EMBEDDING = "TextEmbeddingTask"
    IMAGE_EMBEDDING = "ImageEmbeddingTask"
    WATCH_FOLDERS = "WatchFolderImportTask"

    @staticmethod
    def all():
        return set(item for item in TaskType)



import concurrent
import ctypes
import platform

import datetime
import os
import time
import threading
import numpy as np

from typing import Optional
from concurrent.futures import Future

from sqlmodel import Session, select
from sqlalchemy import func


from .database import DBPriority, VaultDatabase
from .db_models import (
    MetaData,
    Character,
    Picture,
    PictureSet,
    Tag,
    TAG_EMPTY_SENTINEL,
)
from .pixl_logging import get_logger
from .picture_tagger import PictureTagger
from .utils.image_processing.image_utils import ImageUtils
from .tasks.face_quality_task import FaceQualityTask
from .tasks.face_extraction_task import FaceExtractionTask
from .tasks.image_embedding_task import ImageEmbeddingTask
from .tasks.likeness_task import LikenessTask
from .tasks.quality_task import QualityTask
from .tasks.base_task import TaskStatus
from .task_runner import TaskRunner
from .work_planner import WorkPlanner
from .tasks import TaskType

from pixlvault.event_types import EventType


logger = get_logger(__name__)


class Vault:
    AGGRESSIVE_UNLOAD_INTERVAL = 180

    def __enter__(self):
        # Allow use as a context manager for robust cleanup
        return self

    def __exit__(self, _, __, ___):
        self.close()

    """
    Represents a vault for storing images and metadata.

    The vault contains a database that manages a SQLite database and stores the vault description in the metadata table.
    """

    def __init__(
        self,
        image_root: str,
        description: Optional[str] = None,
        server_config_path: Optional[str] = None,
    ):
        """
        Initialize a Vault instance.

        Args:
            db_path (str): Path to the SQLite database file.
            image_root (Optional[str]): Path to the image root directory.
            description (Optional[str]): Description of the vault.
        """
        self.image_root = image_root
        logger.debug(f"Image root: {self.image_root}")
        assert self.image_root is not None, "image_root cannot be None"
        logger.debug(f"Using image_root: {self.image_root}")
        os.makedirs(self.image_root, exist_ok=True)
        assert os.path.exists(self.image_root), (
            f"Image root path does not exist: {self.image_root}"
        )

        self._db_path = os.path.join(self.image_root, "vault.db")
        self.db = VaultDatabase(self._db_path)
        self.set_description(description or "")

        self._picture_tagger = None
        self._last_aggressive_unload_at = 0.0
        self._keep_models_in_memory = True
        self._max_vram_gb = None
        self._server_config_path = server_config_path

        self._planner_watchers = {}
        self._planner_watchers_lock = threading.Lock()
        self._event_listeners = []
        self._event_listeners_lock = threading.Lock()
        self._task_runner = TaskRunner(name="vault-task-runner", num_workers=4)
        self._planner_work_finders = WorkPlanner.work_finders(
            database=self.db,
            picture_tagger_getter=lambda: self._picture_tagger,
            config_path=self._server_config_path,
        )
        self._work_planner = WorkPlanner(
            task_runner=self._task_runner,
            task_finders=list(self._planner_work_finders.values()),
        )
        self._closed = False

        self._task_runner.add_task_complete_callback(self._on_task_completed)
        self._task_runner.add_task_complete_callback(
            self._work_planner.on_task_complete
        )
        self._task_runner.start()
        self._work_planner.start()

    def ensure_ready(self):
        """Initialise the picture tagger so the planner can process work immediately.

        Call this at server startup. Tests that do not need the tagger can skip it;
        tagger init is also triggered lazily by get_worker_future().
        """
        if not self._picture_tagger:
            self._picture_tagger = PictureTagger(image_root=self.image_root)
            self._picture_tagger.set_keep_models_in_memory(self._keep_models_in_memory)
            self._picture_tagger.set_max_vram_usage_gb(self._max_vram_gb)

    def notify(self, event_type: EventType, data=None):
        """
        Notify all relevant workers for a given event type.

        Example:
            vault.notify(Vault.VaultEventType.NEW_PICTURE)
        """
        if self._work_planner and self._work_planner.is_running():
            self._work_planner.wake()
        with self._event_listeners_lock:
            listeners = list(self._event_listeners)
        for listener in listeners:
            try:
                listener(event_type, data)
            except Exception as exc:
                logger.warning("Event listener failed for %s: %s", event_type, exc)

    def add_event_listener(self, listener):
        """Register a callback to be invoked when vault events occur."""
        if not callable(listener):
            raise ValueError("listener must be callable")
        with self._event_listeners_lock:
            if listener not in self._event_listeners:
                self._event_listeners.append(listener)

    def __repr__(self):
        """
        Return a string representation of the Vault instance.

        Returns:
            str: String representation.
        """
        return f"Vault(db_path='{self._db_path}')"

    def close(self):
        """
        Cleanly close the vault, including stopping background workers and closing DB connection.
        """
        if self._closed:
            return
        self._closed = True
        self._work_planner.stop()
        self._task_runner.stop()
        FaceExtractionTask.release_detection_models()
        ImageEmbeddingTask.release_models()
        if self._picture_tagger:
            self._picture_tagger.close()
            del self._picture_tagger
            self._picture_tagger = None
        if self.db:
            self.db.close()
            del self.db
            self.db = None

    def set_keep_models_in_memory(self, keep_models_in_memory: bool):
        previous = self._keep_models_in_memory
        self._keep_models_in_memory = bool(keep_models_in_memory)

        if self._picture_tagger and hasattr(
            self._picture_tagger, "set_keep_models_in_memory"
        ):
            self._picture_tagger.set_keep_models_in_memory(self._keep_models_in_memory)

        if previous and not self._keep_models_in_memory:
            logger.info(
                "keep_models_in_memory disabled; attempting immediate model unload."
            )
            self._last_aggressive_unload_at = 0.0
            progress = self._build_worker_progress_snapshot()
            self._maybe_aggressive_unload(progress)

    def set_max_vram_usage_gb(self, max_vram_gb: Optional[float]):
        self._max_vram_gb = max_vram_gb
        self._task_runner.set_max_vram_usage_gb(max_vram_gb)
        if self._picture_tagger and hasattr(
            self._picture_tagger, "set_max_vram_usage_gb"
        ):
            self._picture_tagger.set_max_vram_usage_gb(max_vram_gb)

    def generate_text_embedding(self, query: str) -> Optional[np.ndarray]:
        """
        Generate a text embedding using PictureTagger.

        Args:
            text (str): Input text to generate embedding for.

        Returns:
            Optional[np.ndarray]: Generated text embedding or None if failed.
        """
        embedding = self._picture_tagger.generate_text_embedding(query=query)
        return embedding[0] if embedding is not None and len(embedding) > 0 else None

    def generate_clip_text_embedding(self, query: str) -> Optional[np.ndarray]:
        """
        Generate a CLIP text embedding for the provided query text.
        """
        return self._picture_tagger.generate_clip_text_embedding(query=query)

    def preprocess_query_words(self, words: list[str]) -> list[str]:
        """
        Preprocess a list of words using the PictureTagger.

        Args:
            words (list[str]): List of input words to preprocess.

        Returns:
            list[str]: Preprocessed list of words.
        """
        preprocessed_words = self._picture_tagger.preprocess_query_words(words=words)
        return preprocessed_words

    def set_description(self, description: str):
        def op(session: Session):
            metadata = session.exec(
                select(MetaData).where(
                    MetaData.schema_version == MetaData.CURRENT_SCHEMA_VERSION
                )
            ).first()
            if metadata is None:
                metadata = MetaData(
                    schema_version=MetaData.CURRENT_SCHEMA_VERSION,
                    description=description,
                )
            else:
                metadata.description = description
            session.add(metadata)
            session.commit()

        self.db.submit_task(op, priority=DBPriority.IMMEDIATE)

    def get_description(self) -> Optional[str]:
        return self.db.submit_task(
            lambda session: (
                session.exec(
                    select(MetaData).where(
                        MetaData.schema_version == MetaData.CURRENT_SCHEMA_VERSION
                    )
                )
                .first()
                .description
            )
        ).result()

    def submit_task(self, task):
        """Submit an in-memory task to the shared task runner."""
        return self._task_runner.submit(task)

    def _on_task_completed(self, task, error):
        if error is not None or task.status != TaskStatus.COMPLETED:
            return

        result = task.result if isinstance(task.result, dict) else {}
        changed = result.get("changed") if isinstance(result, dict) else None
        if not changed:
            return

        if task.type == "TagTask":
            self._notify_worker_ids_processed(TaskType.TAGGER, changed)
            picture_ids = [pic_id for _, pic_id, _, _ in changed]
            if picture_ids:
                self.notify(EventType.CHANGED_TAGS, picture_ids)
            return

        if task.type == "QualityTask":
            self._notify_worker_ids_processed(TaskType.QUALITY, changed)
            self.notify(EventType.QUALITY_UPDATED)
            return

        if task.type == "FaceQualityTask":
            self._notify_worker_ids_processed(TaskType.FACE_QUALITY, changed)
            return

        if task.type == "LikenessParametersTask":
            self._notify_worker_ids_processed(TaskType.LIKENESS_PARAMETERS, changed)
            return

        if task.type == "LikenessTask":
            self._notify_worker_ids_processed(TaskType.LIKENESS, changed)
            return

        if task.type == "FaceExtractionTask":
            self._notify_worker_ids_processed(TaskType.FACE_EXTRACTION, changed)
            picture_ids = result.get("picture_ids") or []
            if picture_ids:
                self.notify(EventType.CHANGED_FACES, picture_ids)
            return

        if task.type == "DescriptionTask":
            self._notify_worker_ids_processed(TaskType.DESCRIPTION, changed)
            self.notify(EventType.CHANGED_DESCRIPTIONS)
            return

        if task.type == "TextEmbeddingTask":
            self._notify_worker_ids_processed(TaskType.TEXT_EMBEDDING, changed)
            return

        if task.type == "ImageEmbeddingTask":
            self._notify_worker_ids_processed(TaskType.IMAGE_EMBEDDING, changed)
            return

        if task.type == "WatchFolderImportTask":
            picture_ids = result.get("imported_picture_ids") or []
            if picture_ids:
                self.notify(EventType.CHANGED_PICTURES, picture_ids)
                self.notify(EventType.PICTURE_IMPORTED, picture_ids)

    def _notify_worker_ids_processed(self, worker_type: TaskType, changed):
        self._notify_planner_ids_processed(worker_type, changed)

    def _watch_planner_id(self, worker_type: TaskType, cls: type, object_id, attr: str):
        future = Future()
        with self._planner_watchers_lock:
            self._planner_watchers[(worker_type, cls, object_id, attr)] = future
        return future

    def _planner_attr_current_value(self, cls: type, object_id: int, attr: str):
        def fetch(session: Session):
            obj = session.get(cls, object_id)
            if obj is None:
                return False, None
            value = getattr(obj, attr, None)
            if attr in {"faces", "hands", "tags"}:
                try:
                    return len(value or []) > 0, value
                except Exception:
                    return False, value
            return value is not None, value

        return self.db.run_immediate_read_task(fetch)

    def _resolve_planner_future_if_already_processed(
        self,
        cls: type,
        object_id: int,
        attr: str,
    ) -> Future | None:
        is_ready, payload = self._planner_attr_current_value(cls, object_id, attr)
        if not is_ready:
            return None
        future = Future()
        future.set_result((object_id, payload))
        return future

    def _notify_planner_ids_processed(self, worker_type: TaskType, changed):
        with self._planner_watchers_lock:
            for cls, object_id, attr, payload in changed:
                future = self._planner_watchers.pop(
                    (worker_type, cls, object_id, attr),
                    None,
                )
                if future:
                    future.set_result((object_id, payload))

    def get_worker_future(
        self, worker_type: TaskType, cls: type, object_id: int, attr: str
    ) -> "concurrent.futures.Future":
        """
        Returns a Future that will be set when the specified worker has processed the given object ID.
        Args:
            worker_type (TaskType): The type of worker to wait for.
        Returns:
            concurrent.futures.Future: Future set to True when completed.
        """
        if not self._picture_tagger:
            self._picture_tagger = PictureTagger(image_root=self.image_root)
            self._picture_tagger.set_keep_models_in_memory(self._keep_models_in_memory)
            self._picture_tagger.set_max_vram_usage_gb(self._max_vram_gb)
        resolved_future = self._resolve_planner_future_if_already_processed(
            cls,
            object_id,
            attr,
        )
        if resolved_future is not None:
            return resolved_future
        return self._watch_planner_id(worker_type, cls, object_id, attr)

    def is_worker_running(self, worker_type: TaskType) -> bool:
        """Check if a specific worker is running."""
        return bool(self._work_planner and self._work_planner.is_running())

    def _is_worker_active(self, worker_type: TaskType) -> bool:
        if not self._work_planner or not self._work_planner.is_running():
            return False
        finder = self._planner_work_finders.get(worker_type)
        if finder is None:
            return False
        finder_name = finder.finder_name()
        return self._work_planner.inflight_count(finder_name) > 0

    def _build_worker_progress_snapshot(self) -> dict:
        progress = {}
        for worker_type in TaskType.all():
            total = int(
                self.db.run_immediate_read_task(self._count_total_pictures) or 0
            )
            if worker_type == TaskType.DESCRIPTION:
                missing = int(
                    self.db.run_immediate_read_task(self._count_missing_descriptions)
                    or 0
                )
                label = "descriptions_generated"
            elif worker_type == TaskType.TAGGER:
                missing = int(
                    self.db.run_immediate_read_task(self._count_missing_tags) or 0
                )
                label = "pictures_tagged"
            elif worker_type == TaskType.QUALITY:
                missing = int(
                    self.db.run_immediate_read_task(self._count_missing_quality) or 0
                )
                label = "quality_scored"
            elif worker_type == TaskType.FACE_QUALITY:
                total = int(
                    self.db.run_immediate_read_task(self._count_total_faces) or 0
                )
                missing = int(
                    self.db.run_immediate_read_task(self._count_missing_face_quality)
                    or 0
                )
                label = "face_quality_scored"
            elif worker_type == TaskType.FACE_EXTRACTION:
                missing = int(
                    self.db.run_immediate_read_task(
                        self._count_missing_feature_extractions
                    )
                    or 0
                )
                label = "features_extracted"
            elif worker_type == TaskType.TEXT_EMBEDDING:
                described = int(
                    self.db.run_immediate_read_task(self._count_total_described) or 0
                )
                missing = int(
                    self.db.run_immediate_read_task(self._count_missing_text_embeddings)
                    or 0
                )
                total = max(described, 0)
                label = "text_embeddings"
            elif worker_type == TaskType.IMAGE_EMBEDDING:
                missing = int(
                    self.db.run_immediate_read_task(
                        self._count_missing_image_embeddings
                    )
                    or 0
                )
                label = "image_embeddings"
            elif worker_type == TaskType.LIKENESS_PARAMETERS:
                missing = int(
                    self.db.run_immediate_read_task(
                        self._count_pending_likeness_parameters
                    )
                    or 0
                )
                label = "likeness_parameters"
            elif worker_type == TaskType.LIKENESS:
                total = int(
                    self.db.run_immediate_read_task(
                        self._count_total_likeness_candidates
                    )
                    or 0
                )
                missing = int(
                    self.db.run_immediate_read_task(self._count_pending_likeness_queue)
                    or 0
                )
                label = "likeness_pairs"
            elif worker_type == TaskType.WATCH_FOLDERS:
                total = 0
                missing = 0
                label = "watch_folder_import"
            else:
                missing = 0
                label = "planner_managed"
            worker_active = self._is_worker_active(worker_type)
            progress[worker_type.value] = {
                "label": label,
                "current": max(total - missing, 0),
                "total": total,
                "remaining": max(missing, 0),
                "updated_at": time.time(),
                "status": "running" if worker_active else "idle",
                "running": worker_active,
                "active": worker_active,
            }
        return progress

    @staticmethod
    def _count_total_pictures(session: Session) -> int:
        result = session.exec(select(func.count()).select_from(Picture)).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_missing_descriptions(session: Session) -> int:
        result = session.exec(
            select(func.count())
            .select_from(Picture)
            .where(Picture.description.is_(None))
        ).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_missing_tags(session: Session) -> int:
        has_real_tag = (Tag.tag.is_not(None)) & (Tag.tag != TAG_EMPTY_SENTINEL)
        result = session.exec(
            select(func.count())
            .select_from(Picture)
            .where(~Picture.tags.any(has_real_tag))
        ).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_missing_feature_extractions(session: Session) -> int:
        result = session.exec(
            select(func.count()).select_from(Picture).where(~Picture.faces.any())
        ).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_missing_quality(session: Session) -> int:
        return QualityTask.count_missing_quality(session)

    @staticmethod
    def _count_total_faces(session: Session) -> int:
        return FaceQualityTask.count_total_faces(session)

    @staticmethod
    def _count_missing_face_quality(session: Session) -> int:
        return FaceQualityTask.count_missing_face_quality(session)

    @staticmethod
    def _count_total_described(session: Session) -> int:
        result = session.exec(
            select(func.count())
            .select_from(Picture)
            .where(Picture.description.is_not(None))
        ).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_missing_text_embeddings(session: Session) -> int:
        result = session.exec(
            select(func.count())
            .select_from(Picture)
            .where(Picture.description.is_not(None))
            .where(Picture.text_embedding.is_(None))
        ).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_missing_image_embeddings(session: Session) -> int:
        return ImageEmbeddingTask.count_remaining(session)

    @staticmethod
    def _count_pending_likeness_parameters(session: Session) -> int:
        result = session.exec(
            select(func.count())
            .select_from(Picture)
            .where(
                (Picture.likeness_parameters.is_(None))
                | (Picture.size_bin_index.is_(None))
            )
        ).one()
        if isinstance(result, (tuple, list)):
            return result[0]
        return result or 0

    @staticmethod
    def _count_pending_likeness_queue(session: Session) -> int:
        return LikenessTask.count_queue(session)

    @staticmethod
    def _count_total_likeness_candidates(session: Session) -> int:
        return LikenessTask.count_total_candidates(session)

    def get_worker_progress(self) -> dict:
        progress = self._build_worker_progress_snapshot()
        self._maybe_aggressive_unload(progress)
        return progress

    def _maybe_aggressive_unload(self, progress: dict):
        if self._keep_models_in_memory:
            return
        if not self._picture_tagger:
            return
        now = time.time()
        if now - self._last_aggressive_unload_at < self.AGGRESSIVE_UNLOAD_INTERVAL:
            return

        any_busy = False
        for snapshot in progress.values():
            status = snapshot.get("status")
            running = bool(snapshot.get("running"))
            if not running:
                continue
            if status in ("idle", "stopped", "uninitialized"):
                continue

            current = int(snapshot.get("current") or 0)
            total = int(snapshot.get("total") or 0)
            remaining = snapshot.get("remaining")
            if remaining is None:
                remaining = max(0, total - current)
            else:
                remaining = max(0, int(remaining))

            has_pending_work = remaining > 0 or (total > 0 and current < total)
            if has_pending_work:
                any_busy = True
                break
        if any_busy:
            return

        logger.warning("All workers idle; aggressively unloading models.")
        try:
            self._picture_tagger.aggressive_unload()
        except Exception as exc:
            logger.warning("Aggressive unload failed for PictureTagger: %s", exc)
        try:
            FaceExtractionTask.release_detection_models()
        except Exception as exc:
            logger.warning(
                "Aggressive unload failed for feature extraction models: %s", exc
            )
        try:
            ImageEmbeddingTask.release_models()
        except Exception as exc:
            logger.warning(
                "Aggressive unload failed for image embedding models: %s", exc
            )
        if platform.system().lower().startswith("linux"):
            try:
                ctypes.CDLL("libc.so.6").malloc_trim(0)
            except Exception:
                pass
        self._last_aggressive_unload_at = now

    def import_default_data(self, add_tagger_test_images: bool = False):
        """
        Import default data into the vault.
        Extend this method to add default pictures or metadata as needed.
        """
        # Add Logo.png to every vault

        logo_src = os.path.join(os.path.dirname(os.path.dirname(__file__)), "Logo.png")
        logo_dest_folder = self.image_root
        logger.debug(f"logo_dest_folder in _import_default_data: {logo_dest_folder}")

        characters = [
            "Esmeralda Vault",
            "Barbara Vault",
            "Barry Vault",
            "Cassandra Vault",
        ]

        def add_character(session: Session, character: Character):
            session.add(character)
            session.commit()
            session.refresh(character)
            char_id = character.id
            char_name = character.name
            # Create reference picture set for this character, using character name
            reference_set = PictureSet(
                name="reference_pictures", description=str(char_name)
            )
            session.add(reference_set)
            session.commit()
            session.refresh(reference_set)
            return char_id, char_name

        for character_name in characters:
            self.db.run_task(
                lambda session: add_character(
                    session,
                    Character(
                        name=character_name, description="Built-in vault character"
                    ),
                ),
                priority=DBPriority.IMMEDIATE,
            )

        picture = ImageUtils.create_picture_from_file(
            image_root_path=logo_dest_folder,
            source_file_path=logo_src,
        )
        picture.description = "PixlVault Logo"
        picture.imported_at = datetime.datetime.now()

        assert picture.file_path

        def add_picture(session: Session, picture: Picture):
            session.add(picture)
            session.commit()
            session.refresh(picture)
            return picture

        picture = self.db.run_task(
            lambda session: add_picture(session, picture),
            priority=DBPriority.IMMEDIATE,
        )

        if add_tagger_test_images:
            # Add all pictures/TaggerTest*.png
            for file in os.listdir(
                os.path.join(os.path.dirname(os.path.dirname(__file__)), "pictures")
            ):
                if file.startswith("TaggerTest") and file.endswith(".png"):
                    src_path = os.path.join(
                        os.path.dirname(os.path.dirname(__file__)),
                        "pictures",
                        file,
                    )
                    pic = ImageUtils.create_picture_from_file(
                        image_root_path=logo_dest_folder,
                        source_file_path=src_path,
                    )
                    pic.description = os.path.basename(src_path)
                    pic.imported_at = datetime.datetime.now()
                    assert pic.file_path
                    self.db.run_task(
                        add_picture,
                        pic,
                        priority=DBPriority.IMMEDIATE,
                    )
                    logger.debug(f"Imported default picture: {pic.file_path}")
        logger.info("Imported default data into the vault.")



from fastapi import APIRouter, Body, HTTPException
from sqlmodel import Session, delete, select

from pixlvault.db_models import (
    Picture,
    Tag,
    TAG_EMPTY_SENTINEL,
)
from pixlvault.event_types import EventType
from pixlvault.pixl_logging import get_logger
from pixlvault.utils import serialize_tag_objects

logger = get_logger(__name__)


def create_router(server) -> APIRouter:
    router = APIRouter()

    @router.post(
        "/pictures/{id}/tags",
        summary="Add tag to picture",
        description="Adds a tag to a picture and removes empty-tag sentinel when appropriate.",
    )
    async def add_tag_to_picture(id: str, payload: dict = Body(...)):
        try:
            try:
                pic_id = int(id)
            except (TypeError, ValueError):
                raise HTTPException(status_code=400, detail="Invalid picture id")
            tag = payload.get("tag")
            if not tag:
                raise HTTPException(status_code=400, detail="Tag is required")

            pic_list = server.vault.db.run_task(
                lambda session: Picture.find(
                    session,
                    id=pic_id,
                    select_fields=["tags"],
                    include_deleted=True,
                    include_unimported=True,
                )
            )
            if not pic_list:
                raise HTTPException(status_code=404, detail="Picture not found")
            pic = pic_list[0]

            existing = next((t for t in pic.tags if t.tag == tag), None)
            if existing is None:

                def update_picture(session, pic_id, tag):
                    pic = Picture.find(
                        session,
                        id=pic_id,
                        select_fields=["tags"],
                        include_deleted=True,
                        include_unimported=True,
                    )[0]
                    sentinel = next(
                        (t for t in pic.tags if t.tag == TAG_EMPTY_SENTINEL),
                        None,
                    )
                    if sentinel is not None:
                        session.delete(sentinel)
                    if not any(t.tag == tag for t in pic.tags):
                        pic.tags.append(Tag(tag=tag, picture_id=pic_id))
                    session.add(pic)
                    session.commit()
                    session.refresh(pic)
                    return pic

                pic = server.vault.db.run_task(update_picture, pic.id, tag)
                server.vault.notify(EventType.CHANGED_TAGS)

            return {"status": "success", "tags": serialize_tag_objects(pic.tags)}
        except HTTPException:
            raise
        except Exception as e:
            logger.error(f"Failed to add tag: {e}")
            raise HTTPException(status_code=500, detail="Failed to add tag")

    @router.get(
        "/pictures/{id}/tags",
        summary="List picture tags",
        description="Returns all tags currently attached to a picture.",
    )
    async def list_picture_tags(id: str):
        try:
            try:
                pic_id = int(id)
            except (TypeError, ValueError):
                raise HTTPException(status_code=400, detail="Invalid picture id")
            pic_list = server.vault.db.run_task(
                lambda session: Picture.find(
                    session,
                    id=pic_id,
                    select_fields=["tags"],
                    include_deleted=True,
                    include_unimported=True,
                )
            )
            if not pic_list:
                raise HTTPException(status_code=404, detail="Picture not found")
            pic = pic_list[0]
            return {
                "id": getattr(pic, "id", None),
                "tags": serialize_tag_objects(pic.tags),
            }
        except HTTPException:
            raise
        except Exception as exc:
            logger.error("Failed to list tags for picture %s: %s", id, exc)
            raise HTTPException(
                status_code=500, detail="Failed to list tags for picture"
            )

    @router.delete(
        "/pictures/{id}/tags/{tag_id}",
        summary="Remove picture tag",
        description="Removes one tag from a picture by numeric tag id and restores empty-tag sentinel when needed.",
    )
    async def remove_tag_from_picture(id: str, tag_id: str):
        try:
            try:
                pic_id = int(id)
            except (TypeError, ValueError):
                raise HTTPException(status_code=400, detail="Invalid picture id")
            if not tag_id.isdigit():
                raise HTTPException(status_code=400, detail="tag_id must be numeric")
            tag_id_int = int(tag_id)

            def update_picture(session, pic_id, tag_id_value):
                pic = Picture.find(
                    session,
                    id=pic_id,
                    select_fields=["tags"],
                    include_deleted=True,
                    include_unimported=True,
                )[0]
                target = session.exec(
                    select(Tag).where(
                        Tag.picture_id == pic_id,
                        Tag.id == tag_id_value,
                    )
                ).first()
                if target is None:
                    raise HTTPException(
                        status_code=404, detail="Tag not found on picture"
                    )
                session.delete(target)
                session.flush()
                remaining = session.exec(
                    select(Tag).where(
                        Tag.picture_id == pic_id,
                        Tag.tag.is_not(None),
                        Tag.tag != TAG_EMPTY_SENTINEL,
                    )
                ).all()
                if not remaining:
                    sentinel = session.exec(
                        select(Tag).where(
                            Tag.picture_id == pic_id,
                            Tag.tag == TAG_EMPTY_SENTINEL,
                        )
                    ).first()
                    if sentinel is None:
                        session.add(Tag(tag=TAG_EMPTY_SENTINEL, picture_id=pic_id))
                session.commit()
                session.refresh(pic)
                return pic

            pic = server.vault.db.run_task(update_picture, pic_id, tag_id_int)
            server.vault.notify(EventType.CHANGED_TAGS)

            return {"status": "success", "tags": serialize_tag_objects(pic.tags)}
        except HTTPException:
            raise
        except Exception as e:
            logger.error(f"Failed to remove tag: {e}")
            raise HTTPException(status_code=500, detail="Failed to remove tag")

    @router.post(
        "/pictures/{id}/tags/remove_all",
        summary="Remove tag everywhere on picture",
        description="Removes a tag value from the picture and its face/hand associations for that picture.",
    )
    async def remove_tag_from_picture_everywhere(id: str, payload: dict = Body(...)):
        try:
            pic_id = int(id)
        except (TypeError, ValueError):
            raise HTTPException(status_code=400, detail="Invalid picture id")
        tag_value = (payload or {}).get("tag")
        if not tag_value:
            raise HTTPException(status_code=400, detail="Tag is required")

        def update_picture(session: Session, pic_id: str, tag_value: str):
            pic_list = Picture.find(
                session,
                id=pic_id,
                select_fields=["tags"],
                include_deleted=True,
                include_unimported=True,
            )
            if not pic_list:
                raise HTTPException(status_code=404, detail="Picture not found")
            pic = pic_list[0]
            tag_ids = [
                t.id for t in pic.tags if t.tag == tag_value and t.id is not None
            ]
            if tag_ids:
                session.exec(delete(Tag).where(Tag.id.in_(tag_ids)))
            session.flush()
            remaining = session.exec(
                select(Tag).where(
                    Tag.picture_id == pic_id,
                    Tag.tag.is_not(None),
                    Tag.tag != TAG_EMPTY_SENTINEL,
                )
            ).all()
            if not remaining:
                sentinel = session.exec(
                    select(Tag).where(
                        Tag.picture_id == pic_id,
                        Tag.tag == TAG_EMPTY_SENTINEL,
                    )
                ).first()
                if sentinel is None:
                    session.add(Tag(tag=TAG_EMPTY_SENTINEL, picture_id=pic_id))
            session.commit()
            session.refresh(pic)
            return pic

        pic = server.vault.db.run_task(update_picture, pic_id, tag_value)
        server.vault.notify(EventType.CHANGED_TAGS)
        return {"status": "success", "tags": serialize_tag_objects(pic.tags)}

    @router.delete(
        "/pictures/{id}/tags",
        summary="Clear all tags on picture",
        description="Removes all tags from a picture in a single operation and restores the empty-tag sentinel.",
    )
    async def clear_all_tags_on_picture(id: str):
        try:
            pic_id = int(id)
        except (TypeError, ValueError):
            raise HTTPException(status_code=400, detail="Invalid picture id")

        def do_clear(session: Session, pic_id: int):
            pic_list = Picture.find(
                session,
                id=pic_id,
                select_fields=["tags"],
                include_deleted=True,
                include_unimported=True,
            )
            if not pic_list:
                raise HTTPException(status_code=404, detail="Picture not found")
            pic = pic_list[0]
            session.exec(delete(Tag).where(Tag.picture_id == pic_id))
            session.add(Tag(tag=TAG_EMPTY_SENTINEL, picture_id=pic_id))
            session.commit()
            session.refresh(pic)
            return pic

        pic = server.vault.db.run_task(do_clear, pic_id)
        server.vault.notify(EventType.CHANGED_TAGS)
        return {"status": "success", "tags": serialize_tag_objects(pic.tags)}

    return router



import json
import re
import secrets
import time
import uuid
from datetime import datetime
from typing import Optional

from fastapi import HTTPException, Request
from fastapi.responses import JSONResponse, Response
from passlib.hash import bcrypt
from pydantic import BaseModel, Field
from sqlmodel import Session, select

from pixlvault.database import DBPriority, VaultDatabase
from pixlvault.db_models import User, UserToken
from pixlvault.utils import default_max_vram_gb


class LoginRequest(BaseModel):
    username: Optional[str] = Field(
        default=None,
        min_length=1,
        description="Username is required",
    )
    password: Optional[str] = Field(
        default=None,
        min_length=8,
        description="Password must be at least 8 characters long",
    )
    token: Optional[str] = Field(
        default=None,
        description="API token for authentication",
    )


class AuthService:
    def __init__(
        self, db: VaultDatabase, server_config: dict, server_config_path: str, logger
    ):
        self._db = db
        self._server_config = server_config
        self._server_config_path = server_config_path
        self._logger = logger
        self.active_session_ids: dict[str, int] = {}
        self.user: Optional[User] = None
        self.password_hash: Optional[str] = None
        self.username: Optional[str] = None
        self._failed_login_attempts: int = 0
        self._login_lockout_until: float = 0.0

    def ensure_secure_when_required(self, request: Request):
        if self._server_config.get("require_ssl", False):
            if request.url.scheme != "https":
                raise HTTPException(
                    status_code=403,
                    detail="HTTPS is required for this operation.",
                )

    def _validate_bcrypt_password_length(self, password: Optional[str]):
        if password is None:
            return
        try:
            byte_length = len(password.encode("utf-8"))
        except Exception:
            byte_length = len(password)
        if byte_length > 72:
            raise HTTPException(
                status_code=400,
                detail=(
                    "Password cannot be longer than 72 bytes. "
                    "Truncate or shorten the password and try again."
                ),
            )

    def get_user(self) -> Optional[User]:
        return self._db.run_task(
            lambda session: session.exec(select(User)).first(),
            priority=DBPriority.IMMEDIATE,
        )

    def ensure_user(self) -> User:
        def ensure_user(session: Session):
            user = session.exec(select(User)).first()
            if user:
                if getattr(user, "max_vram_gb", None) is None:
                    user.max_vram_gb = default_max_vram_gb()
                    session.add(user)
                    session.commit()
                    session.refresh(user)
                return user

            user = User(
                max_vram_gb=default_max_vram_gb(),
            )
            session.add(user)
            session.commit()
            session.refresh(user)
            return user

        user = self._db.run_task(ensure_user, priority=DBPriority.IMMEDIATE)
        self.user = user
        self.password_hash = user.password_hash if user else None
        self.username = user.username if user else None
        return user

    def set_password_hash(self, hashed_password: str):
        def update_user(session: Session):
            user = session.exec(select(User)).first()
            if user is None:
                user = User(max_vram_gb=default_max_vram_gb())
            user.password_hash = hashed_password
            session.add(user)
            session.commit()
            session.refresh(user)
            return user

        user = self._db.run_task(update_user, priority=DBPriority.IMMEDIATE)
        self.password_hash = user.password_hash
        self.user = user
        return user

    def set_username(self, username: str):
        def update_user(session: Session):
            user = session.exec(select(User)).first()
            if user is None:
                user = User(max_vram_gb=default_max_vram_gb())
            user.username = username
            session.add(user)
            session.commit()
            session.refresh(user)
            return user

        user = self._db.run_task(update_user, priority=DBPriority.IMMEDIATE)
        self.username = user.username
        self.user = user
        return user

    def remove_password_hash(self):
        self._logger.info("Removing stored password hash from user database.")

        def clear_user(session: Session):
            user = session.exec(select(User)).first()
            if user is None:
                return None
            user.password_hash = None
            user.username = None
            session.add(user)
            session.commit()
            session.refresh(user)
            return user

        user = self._db.run_task(clear_user, priority=DBPriority.IMMEDIATE)
        self.user = user
        self.password_hash = None
        self.username = None
        self.active_session_ids = {}
        if "PASSWORD_HASH" in self._server_config:
            del self._server_config["PASSWORD_HASH"]
        if "USERNAME" in self._server_config:
            del self._server_config["USERNAME"]
            with open(self._server_config_path, "w") as f:
                json.dump(self._server_config, f, indent=2)
        return user

    def get_user_id(self, request: Request) -> Optional[int]:
        session_id = request.cookies.get("session_id")
        if not session_id:
            return None
        return self.active_session_ids.get(session_id)

    def require_user_id(
        self, request: Request, detail: str = "Not authenticated"
    ) -> int:
        session_id = request.cookies.get("session_id")
        if not session_id:
            raise HTTPException(status_code=401, detail=detail)
        user_id = self.active_session_ids.get(session_id)
        if user_id is None:
            raise HTTPException(status_code=401, detail=detail)
        return user_id

    def get_user_for_request(self, request: Request) -> User:
        user_id = self.require_user_id(request)
        user = self._db.run_task(
            lambda session: session.get(User, user_id),
            priority=DBPriority.IMMEDIATE,
        )
        if user is None:
            raise HTTPException(status_code=404, detail="User not found")
        return user

    def change_password(self, request: Request, payload) -> dict:
        self.ensure_secure_when_required(request)
        user = self.get_user_for_request(request)

        self._validate_bcrypt_password_length(payload.current_password)
        self._validate_bcrypt_password_length(payload.new_password)

        if user.password_hash:
            if not payload.current_password:
                raise HTTPException(
                    status_code=400,
                    detail="Current password is required",
                )
            if not bcrypt.verify(payload.current_password, user.password_hash):
                raise HTTPException(status_code=401, detail="Invalid password")

        hashed_password = bcrypt.hash(payload.new_password)

        def update_user(session: Session, user_id: int):
            db_user = session.get(User, user_id)
            if db_user is None:
                self._logger.debug(
                    "User %s not found in DB when updating",
                    user_id,
                )
                raise HTTPException(
                    status_code=404, detail="User not found when updating"
                )
            db_user.password_hash = hashed_password
            session.add(db_user)
            session.commit()
            session.refresh(db_user)
            return db_user

        updated_user = self._db.run_task(
            update_user, user.id, priority=DBPriority.IMMEDIATE
        )
        self.user = updated_user
        self.password_hash = updated_user.password_hash
        self.username = updated_user.username
        self.active_session_ids = {}
        return {"status": "success"}

    def get_auth_info(self, request: Request) -> dict:
        self.ensure_secure_when_required(request)
        user = self.get_user_for_request(request)
        return {
            "username": user.username,
            "has_password": bool(user.password_hash),
        }

    def create_token(self, request: Request, description: Optional[str]):
        self.ensure_secure_when_required(request)
        user_id = self.require_user_id(request)

        token_value = secrets.token_urlsafe(32)
        token_hash = bcrypt.hash(token_value)

        def create_token(
            session: Session, user_id: int, token_hash: str, desc: Optional[str]
        ):
            user = session.get(User, user_id)
            if user is None:
                raise HTTPException(status_code=404, detail="User not found")
            token = UserToken(
                user_id=user_id,
                token_hash=token_hash,
                created_at=datetime.utcnow(),
                description=desc,
            )
            session.add(token)
            session.commit()
            session.refresh(token)
            return token

        token = self._db.run_task(
            create_token,
            user_id,
            token_hash,
            description,
            priority=DBPriority.IMMEDIATE,
        )

        return {
            "token": token_value,
            "token_id": token.id,
        }

    def list_tokens(self, request: Request):
        self.ensure_secure_when_required(request)
        user_id = self.require_user_id(request)

        def fetch_tokens(session: Session, user_id: int):
            tokens = session.exec(
                select(UserToken)
                .where(UserToken.user_id == user_id)
                .order_by(UserToken.created_at.desc())
            ).all()
            return tokens

        tokens = self._db.run_task(fetch_tokens, user_id, priority=DBPriority.IMMEDIATE)
        return [
            {
                "id": token.id,
                "description": token.description,
                "created_at": token.created_at,
                "last_used_at": token.last_used_at,
            }
            for token in tokens
        ]

    def delete_token(self, request: Request, token_id: int):
        self.ensure_secure_when_required(request)
        user_id = self.require_user_id(request)

        def remove_token(session: Session, user_id: int, token_id: int):
            token = session.get(UserToken, token_id)
            if token is None or token.user_id != user_id:
                raise HTTPException(status_code=404, detail="Token not found")
            session.delete(token)
            session.commit()
            return True

        self._db.run_task(
            remove_token, user_id, token_id, priority=DBPriority.IMMEDIATE
        )
        return {"status": "success", "deleted_id": token_id}

    def check_session(self, request: Request) -> JSONResponse:
        session_id = request.cookies.get("session_id")
        if session_id and session_id in self.active_session_ids:
            return JSONResponse(content={"status": "success"})
        raise HTTPException(status_code=401, detail="Invalid session")

    def login(self, request) -> Response:
        remaining = self._login_lockout_until - time.monotonic()
        if remaining > 0:
            raise HTTPException(
                status_code=429,
                detail="Too many failed login attempts. Try again later.",
                headers={"Retry-After": str(int(remaining) + 1)},
            )
        try:
            response = self._do_login(request)
        except HTTPException as exc:
            if exc.status_code == 401:
                self._failed_login_attempts += 1
                if self._failed_login_attempts >= 5:
                    self._login_lockout_until = time.monotonic() + 60
                    self._logger.warning(
                        "5 failed login attempts — locked out for 60s."
                    )
                else:
                    self._logger.warning(
                        "Login failure #%d.", self._failed_login_attempts
                    )
            raise
        if self._failed_login_attempts:
            self._logger.info(
                "Login succeeded after %d failure(s). Resetting lockout.",
                self._failed_login_attempts,
            )
        self._failed_login_attempts = 0
        self._login_lockout_until = 0.0
        return response

    def _do_login(self, request) -> Response:
        if request.token:
            user = self.get_user()
            if not user:
                raise HTTPException(status_code=404, detail="User not found")

            def fetch_tokens(session: Session, user_id: int):
                tokens = session.exec(
                    select(UserToken).where(UserToken.user_id == user_id)
                ).all()
                return tokens

            tokens = self._db.run_task(
                fetch_tokens, user.id, priority=DBPriority.IMMEDIATE
            )
            matched_token = None
            for token in tokens:
                if bcrypt.verify(request.token, token.token_hash):
                    matched_token = token
                    break
            if matched_token is None:
                raise HTTPException(status_code=401, detail="Invalid token")

            def update_token_last_used(session: Session, token_id: int):
                db_token = session.get(UserToken, token_id)
                if db_token is None:
                    return None
                db_token.last_used_at = datetime.utcnow()
                session.add(db_token)
                session.commit()
                return db_token

            self._db.run_task(
                update_token_last_used,
                matched_token.id,
                priority=DBPriority.IMMEDIATE,
            )

            response = JSONResponse(content={"message": "Login successful."})
        else:
            if not request.username or not request.password:
                raise HTTPException(
                    status_code=400,
                    detail="Username and password are required",
                )

            user = self.get_user() or self.ensure_user()
            if not user.username or not user.password_hash:
                self._validate_bcrypt_password_length(request.password)
                hashed_password = bcrypt.hash(request.password)

                def set_credentials(session: Session):
                    db_user = session.exec(select(User)).first()
                    if db_user is None:
                        db_user = User(max_vram_gb=default_max_vram_gb())
                    db_user.username = request.username
                    db_user.password_hash = hashed_password
                    session.add(db_user)
                    session.commit()
                    session.refresh(db_user)
                    return db_user

                user = self._db.run_task(set_credentials, priority=DBPriority.IMMEDIATE)
                self.user = user
                self.username = user.username
                self.password_hash = user.password_hash
                response = JSONResponse(
                    content={"message": "Username and password set successfully."}
                )
            else:
                if request.username != user.username:
                    raise HTTPException(status_code=401, detail="Invalid username")
                self._validate_bcrypt_password_length(request.password)
                if not bcrypt.verify(request.password, user.password_hash):
                    raise HTTPException(status_code=401, detail="Invalid password")
                response = JSONResponse(content={"message": "Login successful."})

        session_id = str(uuid.uuid4())
        if not user or user.id is None:
            raise HTTPException(status_code=500, detail="User not found")
        self.active_session_ids[session_id] = user.id

        cookie_samesite = self._server_config.get("cookie_samesite", "Lax")
        cookie_secure = self._server_config.get("cookie_secure", False)
        if cookie_samesite == "None" and not cookie_secure:
            self._logger.warning(
                "cookie_samesite=None requires cookie_secure=True for cross-site cookies to work in browsers."
            )
        response.set_cookie(
            key="session_id",
            value=session_id,
            httponly=True,
            samesite=cookie_samesite,
            secure=bool(cookie_secure),
        )
        return response

    def check_registration(self) -> JSONResponse:
        user = self.get_user()
        if not user or not user.username or not user.password_hash:
            return JSONResponse(content={"needs_registration": True})
        return JSONResponse(content={"needs_registration": False})

    def logout(self, response: Response, request: Request):
        session_id = request.cookies.get("session_id")
        if session_id in self.active_session_ids:
            self.active_session_ids.pop(session_id, None)
        response.delete_cookie("session_id", path="/")
        return {"message": "Logged out successfully."}

    async def auth_middleware(
        self, request: Request, call_next, allow_origins, allow_origin_regex
    ):
        excluded_paths = [
            "/login",
            "/docs",
            "/redoc",
            "/openapi.json",
            "/docs/oauth2-redirect",
            "/favicon.ico",
            "/",
            "/version",
            "/check-session",
            "/logout",
        ]
        excluded_prefixes = [
            "/assets/",
            "/pictures/shared/",
            "/docs/",
            "/redoc/",
        ]
        if request.method == "OPTIONS":
            return await call_next(request)

        if request.url.path not in excluded_paths and not any(
            request.url.path.startswith(prefix) for prefix in excluded_prefixes
        ):
            session_id = request.cookies.get("session_id")
            if session_id not in self.active_session_ids:
                self._logger.error(
                    "Invalid session_id. It has expired and the client needs to log in again. When trying to access %s",
                    request.url.path,
                )
                origin = request.headers.get("origin")
                headers = {
                    "Access-Control-Allow-Credentials": "true",
                }
                if origin and (
                    origin in allow_origins
                    or (allow_origin_regex and re.match(allow_origin_regex, origin))
                ):
                    headers["Access-Control-Allow-Origin"] = origin

                return JSONResponse(
                    status_code=401,
                    content={"detail": "Not authenticated"},
                    headers=headers,
                )
        return await call_next(request)
